@@ -31,12 +31,12 @@
     clippy::pedantic,
     clippy::undocumented_unsafe_blocks
 )]
-#![allow(clippy::too_many_lines)]
-#![cfg_attr(feature = "inline-always", allow(clippy::inline_always))]
+#![allow(clippy::inline_always, clippy::single_match_else, clippy::too_many_lines)]
 
 // This crate should work on targets with power-of-two pointer widths,
 // but it is not clear how it will work on targets without them.
 // There are currently no 8-bit, 128-bit, or higher builtin targets.
+// Note that Rust (and C99) pointers must be at least 16-bits: https://github.com/rust-lang/rust/pull/49305
 #[cfg(not(any(
     target_pointer_width = "16",
     target_pointer_width = "32",
@@ -44,7 +44,7 @@
 )))]
 compile_error!(
     "atomic-memcpy currently only supports targets with {16,32,64}-bit pointer width; \
-     if you need support for 8-bit or others, \
+     if you need support for others, \
      please submit an issue at <https://github.com/taiki-e/atomic-memcpy>"
 );
 
@@ -197,11 +197,9 @@ pub unsafe fn atomic_store<T>(dst: *mut T, val: T, order: Ordering) {
 /// Note that the use of `--cfg atomic_memcpy_unsafe_volatile` is
 /// undefined behavior in the multi-threaded environment, since volatile
 /// read/write does not guarantee anything about data race.
-#[cfg(not(any(
-    target_pointer_width = "16",
-    all(target_arch = "riscv32", not(target_feature = "a"), target_os = "none"),
-    atomic_memcpy_unsafe_volatile,
-)))]
+#[cfg(not(any(target_pointer_width = "16", atomic_memcpy_unsafe_volatile)))]
+// Note: riscv_target_feature is not stable, so `target_feature = "a"` does not work on stable.
+#[cfg_attr(all(target_arch = "riscv32", target_os = "none"), cfg(target_has_atomic = "ptr"))]
 mod imp {
     #[cfg(not(target_pointer_width = "16"))]
     use core::sync::atomic::AtomicU32;
@@ -213,16 +211,12 @@ mod imp {
         sync::atomic::{AtomicU16, AtomicUsize, Ordering},
     };
 
-    #[cfg(target_pointer_width = "32")]
-    type Half = u16;
-    #[cfg(target_pointer_width = "32")]
-    type AtomicHalf = AtomicU16;
-
-    #[cfg(target_pointer_width = "64")]
-    type Half = u32;
-    #[cfg(target_pointer_width = "64")]
-    type AtomicHalf = AtomicU32;
-
+    // Boundary to make the fields of LoadState private.
+    //
+    // Note that this is not a complete safe/unsafe boundary[1], since it is still
+    // possible to pass an invalid pointer to the constructor.
+    //
+    // [1]: https://www.ralfj.de/blog/2016/01/09/the-scope-of-unsafe.html
     mod load {
         use core::{
             mem,
@@ -307,9 +301,16 @@ mod imp {
                 }
             }
 
+            /// Note: The remaining bytes smaller than usize are ignored.
+            ///
+            /// # Safety
+            ///
+            /// - `self.src` must be properly aligned for `usize`.
+            ///
+            /// There is no alignment requirement for `self.result`.
             #[cfg_attr(feature = "inline-always", inline(always))]
             #[cfg_attr(not(feature = "inline-always"), inline)]
-            pub(super) unsafe fn atomic_load_usize(&mut self) {
+            pub(super) unsafe fn atomic_load_usize_to_end(&mut self) {
                 while self.remaining() >= mem::size_of::<usize>() {
                     // SAFETY:
                     // - the caller must guarantee that `src` is properly aligned for `usize`.
@@ -326,27 +327,6 @@ mod imp {
                     }
                 }
             }
-
-            #[cfg_attr(feature = "inline-always", inline(always))]
-            #[cfg_attr(not(feature = "inline-always"), inline)]
-            #[cfg(not(target_pointer_width = "16"))]
-            pub(super) unsafe fn atomic_load_half(&mut self) {
-                use super::{AtomicHalf, Half};
-                debug_assert!(self.remaining() >= mem::size_of::<Half>());
-                // SAFETY:
-                // - the caller must guarantee that both `src` and `dst` are properly aligned for `Half`.
-                // - the caller must guarantee that the remaining bytes is greater than
-                //   or equal to `size_of::<Half>()`.
-                // Therefore, due to `LoadState`'s invariant:
-                // - `src` is valid to atomic read of `Half`.
-                // - `result` is valid to write of `Half`.
-                unsafe {
-                    let val = self.src::<AtomicHalf>().load(Ordering::Relaxed);
-                    self.result::<Half>().write(val);
-                    // SAFETY: we've filled `size_of::<Half>()` bytes.
-                    self.advance(mem::size_of::<Half>());
-                }
-            }
         }
     }
 
@@ -423,38 +403,6 @@ mod imp {
             return result;
         }
 
-        // HACK: Miri cannot track uninitialized bytes on a per byte basis for
-        // partially initialized scalars: https://github.com/rust-lang/rust/issues/69488
-        //
-        // This hack allows miri to properly track the use of uninitialized
-        // bytes. See also tests/uninit.rs that is a test to check if
-        // valgrind/sanitizer/miri can properly detect the use of uninitialized
-        // bytes.
-        //
-        // Note: With or without this hack, atomic load/store of integers
-        // containing uninitialized bytes is technically an undefined behavior.
-        // The only purpose of this hack is to make sure that Miri errors for
-        // atomic load/store of integers containing uninitialized bytes
-        // (which is probably not a problem and uncharted territory at best [1] [2] [3],
-        // and can be detected by `-Zmiri-check-number-validity` [4]),
-        // do not mask Miri errors for the use of uninitialized bytes (which is definitely a problem).
-        // See also tests/padding.rs.
-        //
-        // [1] https://github.com/crossbeam-rs/crossbeam/issues/315
-        // [2] https://github.com/rust-lang/unsafe-code-guidelines/issues/158
-        // [3] https://github.com/rust-lang/unsafe-code-guidelines/issues/71
-        // [4] https://github.com/rust-lang/miri/pull/1904
-        //
-        // rust-lang/rust#69488 affects only CTFE(compile-time function evaluation)/Miri
-        // and atomic operations cannot be called in const context, so our code
-        // is only affected in the case of cfg(miri).
-        if cfg!(miri) {
-            let mut state = load::LoadState::new(result.as_mut_ptr(), src);
-            state.atomic_load_u8(state.remaining());
-            debug_assert_eq!(state.remaining(), 0);
-            return result;
-        }
-
         // Branch 1: If the alignment of `T` is less than usize, but `T` can be read as
         // at least one or more usize, compute the align offset and read it
         // like `(&[AtomicU8], &[AtomicUsize], &[AtomicU8])`.
@@ -462,56 +410,6 @@ mod imp {
             && mem::size_of::<T>() >= mem::size_of::<usize>() * 4
         {
             let mut state = load::LoadState::new(result.as_mut_ptr(), src);
-            #[cfg(not(target_pointer_width = "16"))]
-            {
-                // Since the caller guarantees that the pointer is properly aligned,
-                // if `T` has an alignment of half of usize, there are only two
-                // patterns: read as usize from the first byte, or read as usize
-                // after reading `usize / 2` bytes.
-                //
-                // on 64-bit:
-                // | 8 | 8 | 8 | 4 |
-                // | 4 | 8 | 8 | 8 |
-                // or
-                // | 8 | 8 | 8 |
-                // | 4 | 8 | 8 | 4 |
-                //
-                // Handling this case manually can reduce the number of instructions
-                // significantly compared to using align_offset.
-                if mem::align_of::<T>() >= mem::align_of::<Half>() {
-                    if src as usize % mem::size_of::<usize>() == 0 {
-                        // SAFETY:
-                        // - we've checked that `src` is properly aligned for `usize`.
-                        // - the remaining bytes is greater than or equal to `size_of::<usize>()`.
-                        unsafe { state.atomic_load_usize() }
-                    } else {
-                        debug_assert_eq!(
-                            src as usize % mem::size_of::<usize>(),
-                            mem::size_of::<Half>()
-                        );
-                        // SAFETY:
-                        // - the caller must guarantee that `src` is properly aligned for `T`.
-                        // - `T` has an alignment greater than or equal to `Half`.
-                        // - the remaining bytes is greater than or equal to `size_of::<usize>()`.
-                        unsafe {
-                            state.atomic_load_half();
-                            // SAFETY: we've advanced `size_of::<Half>()` bytes,
-                            // so now `state.src` is definitely aligned.
-                            state.atomic_load_usize();
-                        }
-                    }
-                    // Load remaining bytes.
-                    if state.remaining() != 0 {
-                        debug_assert_eq!(state.remaining(), mem::size_of::<Half>());
-                        // SAFETY:
-                        // - the caller must guarantee that `src` is properly aligned for `T`.
-                        // - `T` has an alignment greater than or equal to `Half`.
-                        // - the remaining bytes is equal to `size_of::<Half>()`.
-                        unsafe { state.atomic_load_half() }
-                    }
-                    return result;
-                }
-            }
             let offset = (src as *const u8).align_offset(mem::align_of::<AtomicUsize>());
             // Note: align_offset may returns usize::MAX: https://github.com/rust-lang/rust/issues/62420
             if state.remaining() >= offset {
@@ -525,9 +423,9 @@ mod imp {
                 //   or equal to `size_of::<usize>()`.
                 //
                 // In this branch, the pointer to `state.result` is usually
-                // not properly aligned, so we use `atomic_load_usize_result_unaligned`,
+                // not properly aligned, so we use `atomic_load_usize_to_end`,
                 // which has no requirement for alignment of `state.result`.
-                unsafe { state.atomic_load_usize() }
+                unsafe { state.atomic_load_usize_to_end() }
                 // Load remaining bytes per byte.
                 state.atomic_load_u8(state.remaining());
                 debug_assert_eq!(state.remaining(), 0);
@@ -599,6 +497,10 @@ mod imp {
         result
     }
 
+    // Boundary to make the fields of StoreState private.
+    //
+    // Note that this is not a complete safe/unsafe boundary, since it is still
+    // possible to pass an invalid pointer to the constructor.
     mod store {
         use core::{
             mem,
@@ -682,9 +584,16 @@ mod imp {
                 }
             }
 
+            /// Note: The remaining bytes smaller than usize are ignored.
+            ///
+            /// # Safety
+            ///
+            /// - `self.dst` must be properly aligned for `usize`.
+            ///
+            /// There is no alignment requirement for `self.src`.
             #[cfg_attr(feature = "inline-always", inline(always))]
             #[cfg_attr(not(feature = "inline-always"), inline)]
-            pub(super) unsafe fn atomic_store_usize(&mut self) {
+            pub(super) unsafe fn atomic_store_usize_to_end(&mut self) {
                 while self.remaining() >= mem::size_of::<usize>() {
                     // SAFETY:
                     // - the caller must guarantee that `dst` is properly aligned for `usize`.
@@ -701,27 +610,6 @@ mod imp {
                     }
                 }
             }
-
-            #[cfg_attr(feature = "inline-always", inline(always))]
-            #[cfg_attr(not(feature = "inline-always"), inline)]
-            #[cfg(not(target_pointer_width = "16"))]
-            pub(super) unsafe fn atomic_store_half(&mut self) {
-                use super::{AtomicHalf, Half};
-                debug_assert!(self.remaining() >= mem::size_of::<Half>());
-                // SAFETY:
-                // - the caller must guarantee that `dst` is properly aligned for `Half`.
-                // - we've checked that the remaining bytes is greater than
-                //   or equal to `size_of::<Half>()`.
-                // Therefore, due to `StoreState`'s invariant:
-                // - `src` is valid to read of `Half`.
-                // - `dst` is valid to atomic write of `Half`.
-                unsafe {
-                    let val = self.src::<Half>().read();
-                    self.dst::<AtomicHalf>().store(val, Ordering::Relaxed);
-                    // SAFETY: we've filled `size_of::<Half>()` bytes.
-                    self.advance(mem::size_of::<Half>());
-                }
-            }
         }
     }
 
@@ -769,15 +657,6 @@ mod imp {
             return;
         }
 
-        // HACK: See the `atomic_load` function for the detailed comment.
-        if cfg!(miri) {
-            let mut state = store::StoreState::new(dst, &*val);
-            state.atomic_store_u8(state.remaining());
-            debug_assert_eq!(state.remaining(), 0);
-            mem::forget(guard);
-            return;
-        }
-
         // Branch 1: If the alignment of `T` is less than usize, but `T` can be write as
         // at least one or more usize, compute the align offset and write it
         // like `(&[AtomicU8], &[AtomicUsize], &[AtomicU8])`.
@@ -785,44 +664,6 @@ mod imp {
             && mem::size_of::<T>() >= mem::size_of::<usize>() * 4
         {
             let mut state = store::StoreState::new(dst, &*val);
-            #[cfg(not(target_pointer_width = "16"))]
-            {
-                // See the `atomic_load` function for the detailed comment.
-                if mem::align_of::<T>() >= mem::align_of::<Half>() {
-                    if dst as usize % mem::size_of::<usize>() == 0 {
-                        // SAFETY:
-                        // - we've checked that `dst` is properly aligned for `usize`.
-                        // - the remaining bytes is greater than or equal to `size_of::<usize>()`.
-                        unsafe { state.atomic_store_usize() }
-                    } else {
-                        debug_assert_eq!(
-                            dst as usize % mem::size_of::<usize>(),
-                            mem::size_of::<Half>()
-                        );
-                        // SAFETY:
-                        // - the caller must guarantee that `dst` is properly aligned for `T`.
-                        // - `T` has an alignment greater than or equal to `Half`.
-                        // - the remaining bytes is greater than or equal to `size_of::<usize>()`.
-                        unsafe {
-                            state.atomic_store_half();
-                            // SAFETY: we've advanced `size_of::<Half>()` bytes,
-                            // so now `state.dst` is definitely aligned.
-                            state.atomic_store_usize();
-                        }
-                    }
-                    // Store remaining bytes.
-                    if state.remaining() != 0 {
-                        debug_assert_eq!(state.remaining(), mem::size_of::<Half>());
-                        // SAFETY:
-                        // - the caller must guarantee that `src` is properly aligned for `T`.
-                        // - `T` has an alignment greater than or equal to `Half`.
-                        // - the remaining bytes is equal to `size_of::<Half>()`.
-                        unsafe { state.atomic_store_half() }
-                    }
-                    mem::forget(guard);
-                    return;
-                }
-            }
             let offset = (dst as *mut u8).align_offset(mem::align_of::<AtomicUsize>());
             // Note: align_offset may returns usize::MAX: https://github.com/rust-lang/rust/issues/62420
             if state.remaining() >= offset {
@@ -836,10 +677,10 @@ mod imp {
                 //   or equal to `size_of::<usize>()`.
                 //
                 // In this branch, the pointer to `state.src` is usually
-                // not properly aligned, so we use `atomic_store_usize_src_unaligned`,
+                // not properly aligned, so we use `atomic_store_usize_to_end`,
                 // which has no requirement for alignment of `state.src`.
                 unsafe {
-                    state.atomic_store_usize();
+                    state.atomic_store_usize_to_end();
                 }
                 // Store remaining bytes per byte.
                 state.atomic_store_u8(state.remaining());
@@ -937,18 +778,18 @@ mod imp {
         r
     }
 
-    /// Atomic integers larger than the pointer size often does not have the
-    /// same alignment as the corresponding integer types.
-    ///
-    /// ```console
-    /// $ rustc --print cfg --target x86_64-apple-darwin | grep -E 'target_has_atomic_.*(64|128)'
-    /// target_has_atomic_equal_alignment="64"
-    /// target_has_atomic_load_store="128"
-    /// target_has_atomic_load_store="64"
-    /// ```
-    ///
-    /// It's unlikely that the same thing will happen with an atomic type
-    /// less than or equal to the pointer size, but we'll check just in case.
+    // Atomic integers larger than the pointer size often does not have the
+    // same alignment as the corresponding integer types.
+    //
+    // ```console
+    // $ rustc --print cfg --target x86_64-apple-darwin | grep -E 'target_has_atomic_.*(64|128)'
+    // target_has_atomic_equal_alignment="64"
+    // target_has_atomic_load_store="128"
+    // target_has_atomic_load_store="64"
+    // ```
+    //
+    // It's unlikely that the same thing will happen with an atomic type
+    // less than or equal to the pointer size, but we'll check just in case.
     #[cfg_attr(feature = "inline-always", inline(always))]
     #[cfg_attr(not(feature = "inline-always"), inline)]
     fn static_assert_atomic_alignment() {
@@ -973,9 +814,10 @@ mod imp {
 
 #[cfg(any(
     target_pointer_width = "16",
-    all(target_arch = "riscv32", not(target_feature = "a"), target_os = "none"),
-    atomic_memcpy_unsafe_volatile,
+    all(target_arch = "riscv32", target_os = "none"),
+    atomic_memcpy_unsafe_volatile
 ))]
+#[cfg_attr(all(target_arch = "riscv32", target_os = "none"), cfg(not(target_has_atomic = "ptr")))]
 mod imp {
     #[cfg_attr(feature = "inline-always", inline(always))]
     #[cfg_attr(not(feature = "inline-always"), inline)]
